import 'dart:typed_data';

import 'package:chacha20/chacha20.dart';
import 'package:test/test.dart';

void main() {
  group("Chacha20:", () {
    test("randomKey(): two results are not equal", () {
      final key = Chacha20.randomKey();
      expect(key, hasLength(32));
      expect(key, isNot(orderedEquals(Chacha20.randomKey())));
    });

    test("randomNonce(): two results are not equal", () {
      final nonce = Chacha20.randomNonce();
      expect(nonce, hasLength(12));
      expect(nonce, isNot(orderedEquals(Chacha20.randomNonce())));
    });

    test("initialize(...): throws ArgumentError when 'key' is null", () {
      final converter = new Chacha20();
      expect(() {
        converter.initialize(key: null, nonce: new Uint8List(12));
      }, throwsA(const TypeMatcher<ArgumentError>()));
      expect(converter.initialState[0], equals(0));
    });

    test("initialize(...): throws ArgumentError when 'key' has a wrong length",
        () {
      final converter = new Chacha20();
      expect(() {
        converter.initialize(
          key: new List<int>(31),
          nonce: new Uint8List(12),
        );
      }, throwsA(const TypeMatcher<ArgumentError>()));
      expect(converter.initialState[0], equals(0));
    });

    test(
        "initialize(...): throws ArgumentError when 'nonce' has a wrong length",
        () {
      final converter = new Chacha20();
      expect(() {
        converter.initialize(
            key: new List<int>.generate(32, (i) => i),
            nonce: new List<int>(11));
      }, throwsA(const TypeMatcher<ArgumentError>()));
      expect(converter.initialState[0], equals(0));
    });

    test("initialize(...): works with null nonce", () {
      final converter = new Chacha20();
      converter.initialize(
        key: new List<int>.generate(32, (i) => i),
        nonce: new List<int>.generate(12, (i) => i),
      );
      expect(converter.initialState.sublist(13, 16).every((item) => item == 0),
          isFalse);
      converter.initialize(
        key: new List<int>.generate(32, (i) => i),
        nonce: null,
      );
      expect(converter.initialState.sublist(13, 16).every((item) => item == 0),
          isTrue);
    });

    test("incrementNonce(): [..., 0,0,2] --> [...0,0,3]", () {
      var nonce = new Uint8List(12);
      nonce[11] = 2;
      nonce = Chacha20.incrementNonceBytes(nonce);
      expect(nonce[11], equals(3));
      expect(nonce[10], equals(0));
      expect(nonce[9], equals(0));
    });

    test("incrementNonce(): [..., 0,2,255] --> [...0,3,0]", () {
      var nonce = new Uint8List(12);
      nonce[11] = 255;
      nonce[10] = 2;
      nonce = Chacha20.incrementNonceBytes(nonce);
      expect(nonce[11], equals(0));
      expect(nonce[10], equals(3));
      expect(nonce[9], equals(0));
    });

    test("incrementNonce(): [..., 2,255,255] --> [...3,0,0]", () {
      var nonce = new Uint8List(12);
      nonce[11] = 255;
      nonce[10] = 255;
      nonce[9] = 2;
      nonce = Chacha20.incrementNonceBytes(nonce);
      expect(nonce[11], equals(0));
      expect(nonce[10], equals(0));
      expect(nonce[9], equals(3));
    });

    test("keyStreamIndex: can be mutated", () {
      final converter = new Chacha20();
      converter.initialize(key: new Uint8List(32), nonce: new Uint8List(12));
      expect(converter.keyStreamIndex, equals(0));

      converter.keyStreamIndex = 1;
      expect(converter.keyStreamIndex, equals(1));
      expect(converter.initialState[12], equals(0));

      converter.keyStreamIndex = 128;
      expect(converter.keyStreamIndex, equals(128));
      expect(converter.initialState[12], equals(2));

      converter.keyStreamIndex = 129;
      expect(converter.keyStreamIndex, equals(129));
      expect(converter.initialState[12], equals(2));

      converter.keyStreamIndex = 64;
      expect(converter.keyStreamIndex, equals(64));
      expect(converter.initialState[12], equals(1));
    });

    test("keyStreamIndex: is automatically incremented", () {
      final converter = new Chacha20();
      converter.initialize(key: new Uint8List(32), nonce: new Uint8List(12));
      expect(converter.keyStreamIndex, equals(0));

      // 0 --> 0
      converter.convert([]);
      expect(converter.keyStreamIndex, equals(0));

      // 0 --> 3
      converter.convert([1, 2, 3]);
      expect(converter.keyStreamIndex, equals(3));
      expect(converter.initialState[12], equals(0));

      // 3 --> 63
      converter.convert(new List.filled(60, 0));
      expect(converter.keyStreamIndex, equals(63));
      expect(converter.initialState[12], equals(0));

      // 63 -> 64
      converter.convert([1]);
      expect(converter.keyStreamIndex, equals(64));
      expect(converter.initialState[12], equals(1));
    });

    test("convert(...): throws StateError if state is invalid", () {
      final converter = new Chacha20();
      expect(
        () => converter.convert(<int>[]),
        throwsA(const TypeMatcher<StateError>()),
      );
    });

    test(
        "convert(...): throws StateError if state is invalid (initialize(), clear())",
        () {
      final converter = new Chacha20();
      converter.initialize(key: new Uint8List(32), nonce: new Uint8List(12));
      converter.clear();
      expect(
        () => converter.convert(<int>[]),
        throwsA(const TypeMatcher<StateError>()),
      );
    });

    test(
        "convert(...): succeeds after initialize(...); clear(); initialize(...)",
        () {
      final converter = new Chacha20();
      converter.initialize(key: new Uint8List(32), nonce: new Uint8List(12));
      converter.clear();
      converter.initialize(key: new Uint8List(32), nonce: new Uint8List(12));
      converter.convert(<int>[]);
    });

    test("convert(...): all input lengths 0...1000 behave consistently", () {
      final converter = new Chacha20();
      converter.initialize(key: new Uint8List(32), nonce: new Uint8List(12));

      // Clear text is a sequence of 'a' letters
      final clearText = new Uint8List(1000);
      final charCodeForA = "a".codeUnitAt(0);
      clearText.fillRange(0, clearText.length, charCodeForA);

      for (var n = 0; n < 1000; n++) {
        final clearTextSection = new Uint8List.view(clearText.buffer, clearText.offsetInBytes, n);
        // Encrypt
        converter.keyStreamIndex = 0;
        final encrypted = converter.convert(clearTextSection);

        // Decrypt
        converter.keyStreamIndex = 0;
        final decrypted = converter.convert(encrypted);
        expect(decrypted.length, equals(n));

        // Test that the decrypted matches clear text.
        expect(
          decrypted,
          clearTextSection,
        );
      }
    });

    test(
        "convert(...): various 'keyStreamIndex' values and input lengths behave consistently",
        () async {
      final converter = new Chacha20();
      converter.initialize(key: new Uint8List(32), nonce: new Uint8List(12));
      final input = new List.filled(130, "a".codeUnitAt(0));
      final expectedOutput = converter.convert(input);

      for (var skip in const [0, 1, 63, 64, 65, 127, 128, 129]) {
        for (var take in [
          0,
          1,
          63,
          64,
          65,
          127,
          128,
          129,
          input.length - skip
        ]) {
          // If 'take' is too large
          if (take > input.length - skip) {
            // Go to next case
            continue;
          }

          // Take slice of input bytes and encrypted bytes
          final inputSlice = input.skip(skip).take(take).toList();
          final expectedOutputSlice =
              expectedOutput.skip(skip).take(take).toList();

          // Test converting the slice
          converter.keyStreamIndex = skip;
          expect(
            converter.convert(inputSlice),
            orderedEquals(expectedOutputSlice),
          );

          expect(
            converter.keyStreamIndex,
            equals(skip + take),
          );
          expect(
            converter.initialState[12],
            equals((skip + take) ~/ 64),
          );

          // Test converting the remaining bytes
          expect(
            converter.convert(input.skip(skip + take).toList()),
            orderedEquals(expectedOutput.skip(skip + take).toList()),
          );
          expect(converter.keyStreamIndex, equals(130));
          expect(converter.initialState[12], equals(2));
        }
      }
    });

    test("RFC 7539: quarter round example", () {
      /// -----------------------------------
      /// These constants are from RFC 7539:
      /// https://tools.ietf.org/html/rfc7539
      /// -----------------------------------
      final state =
          _uint32ListFromHex("""11111111 01020304 9b8d6f43 01234567""");

      final expectedState =
          _uint32ListFromHex("""ea2a92f4 cb1cf8ce 4581472e 5881c4bb""");

      /// ------------------------------
      /// End of constants from RFC 7539
      /// ------------------------------

      new NonInlinedChacha20().quarterRound(state, 0, 1, 2, 3);

      expect(_hexFrom(state), equals(_hexFrom(expectedState)));
    });

    test("RFC 7539: quarter round on the state example", () {
      /// -----------------------------------
      /// These constants are from RFC 7539:
      /// https://tools.ietf.org/html/rfc7539
      /// -----------------------------------

      final state = _uint32ListFromHex("""
        879531e0  c5ecf37d  516461b1  c9a62f8a
        44c20ef3  3390af7f  d9fc690b  2a5f714c
        53372767  b00a5631  974c541a  359e9963
        5c971061  3d631689  2098d9d6  91dbd320
      """);

      final expectedState = _uint32ListFromHex("""
        879531e0  c5ecf37d  bdb886dc  c9a62f8a
        44c20ef3  3390af7f  d9fc690b  cfacafd2
        e46bea80  b00a5631  974c541a  359e9963
        5c971061  ccc07c79  2098d9d6  91dbd320
      """);

      /// ------------------------------
      /// End of constants from RFC 7539
      /// ------------------------------

      new NonInlinedChacha20().quarterRound(state, 2, 7, 8, 13);

      expect(_hexFrom(state), equals(_hexFrom(expectedState)));
    });

    group("RFC 7539: block function example", () {
      /// -----------------------------------
      /// These constants are from RFC 7539:
      /// https://tools.ietf.org/html/rfc7539
      /// -----------------------------------

      final key = _bytesFromHex(
        "00:01:02:03:04:05:06:07:08:09:0a:0b:0c:0d:0e:0f:10:11:12:13:14:15:16:17:18:19:1a:1b:1c:1d:1e:1f",
      );

      final nonce = _bytesFromHex("00:00:00:09:00:00:00:4a:00:00:00:00");

      final initialKeyStreamIndex = 64;

      final expectedState = _uint32ListFromHex("""
        61707865  3320646e  79622d32  6b206574
        03020100  07060504  0b0a0908  0f0e0d0c
        13121110  17161514  1b1a1918  1f1e1d1c
        00000001  09000000  4a000000  00000000
      """);

      final expectedStateAfterEncrypting = _uint32ListFromHex("""
        e4e7f110  15593bd1  1fdd0f50  c47120a3
        c7f4d1c7  0368c033  9aaa2204  4e6cd4c3
        466482d2  09aa9f07  05d7c214  a2028bd9
        d19c12b5  b94e16de  e883d0cb  4e3c50a2
      """);

      /// ------------------------------
      /// End of constants from RFC 7539
      /// ------------------------------

      final converter = new Chacha20();
      converter.initialize(
        key: key,
        nonce: nonce,
        keyStreamIndex: initialKeyStreamIndex,
      );

      test("initialization", () {
        final state = converter.initialState;
        expect(_hexFrom(state), equals(_hexFrom(expectedState)));
      });

      test("mixing", () {
        final state = new Uint32List(16);
        converter.encryptState(state);
        expect(_hexFrom(state), equals(_hexFrom(expectedStateAfterEncrypting)));
      });
    });

    group("RFC 7539: encryption example", () {
      /// -----------------------------------
      /// These constants are from RFC 7539:
      /// https://tools.ietf.org/html/rfc7539
      /// -----------------------------------

      final cleartext =
          "Ladies and Gentlemen of the class of '99: If I could offer you only one tip for the future, sunscreen would be it."
              .runes
              .toList();

      final key = _bytesFromHex(
        "00:01:02:03:04:05:06:07:08:09:0a:0b:0c:0d:0e:0f:10:11:12:13:14:15:16:17:18:19:1a:1b:1c:1d:1e:1f",
      );

      final nonce = _bytesFromHex("00:00:00:00:00:00:00:4a:00:00:00:00");

      final initialKeyStreamIndex = 64;

      final expectedFirstBlockInitialization = _uint32ListFromHex("""
        61707865  3320646e  79622d32  6b206574
        03020100  07060504  0b0a0908  0f0e0d0c
        13121110  17161514  1b1a1918  1f1e1d1c
        00000001  00000000  4a000000  00000000
      """);

      final expectedSecondBlockInitialization = _uint32ListFromHex("""
        61707865  3320646e  79622d32  6b206574
        03020100  07060504  0b0a0908  0f0e0d0c
        13121110  17161514  1b1a1918  1f1e1d1c
        00000002  00000000  4a000000  00000000
      """);

      final expectedFirstStateEncrypted = _uint32ListFromHex("""
        f3514f22  e1d91b40  6f27de2f  ed1d63b8
        821f138c  e2062c3d  ecca4f7e  78cff39e
        a30a3b8a  920a6072  cd7479b5  34932bed
        40ba4c79  cd343ec6  4c2c21ea  b7417df0
      """);

      final expectedSecondStateEncrypted = _uint32ListFromHex("""
        9f74a669  410f633f  28feca22  7ec44dec
        6d34d426  738cb970  3ac5e9f3  45590cc4
        da6e8b39  892c831a  cdea67c1  2b7e1d90
        037463f3  a11a2073  e8bcfb88  edc49139
      """);

      final expectedKeyStream = _bytesFromHex("""
        22:4f:51:f3:40:1b:d9:e1:2f:de:27:6f:b8:63:1d:ed:8c:13:1f:82:3d:2c:06
        e2:7e:4f:ca:ec:9e:f3:cf:78:8a:3b:0a:a3:72:60:0a:92:b5:79:74:cd:ed:2b
        93:34:79:4c:ba:40:c6:3e:34:cd:ea:21:2c:4c:f0:7d:41:b7:69:a6:74:9f:3f
        63:0f:41:22:ca:fe:28:ec:4d:c4:7e:26:d4:34:6d:70:b9:8c:73:f3:e9:c5:3a
        c4:0c:59:45:39:8b:6e:da:1a:83:2c:89:c1:67:ea:cd:90:1d:7e:2b:f3:63              
      """);

      final expectedEncrypted = _bytesFromHex("""
        6e 2e 35 9a 25 68 f9 80 41 ba 07 28 dd 0d 69 81
        e9 7e 7a ec 1d 43 60 c2 0a 27 af cc fd 9f ae 0b
        f9 1b 65 c5 52 47 33 ab 8f 59 3d ab cd 62 b3 57
        16 39 d6 24 e6 51 52 ab 8f 53 0c 35 9f 08 61 d8
        07 ca 0d bf 50 0d 6a 61 56 a3 8e 08 8a 22 b6 5e
        52 bc 51 4d 16 cc f8 06 81 8c e9 1a b7 79 37 36
        5a f9 0b bf 74 a3 5b e6 b4 0b 8e ed f2 78 5e 42
        87 4d                   
      """);

      /// ------------------------------
      /// End of constants from RFC 7539
      /// ------------------------------

      Chacha20 converter;
      setUp(() {
        converter = new Chacha20();
        converter.initialize(
          key: key,
          nonce: nonce,
          keyStreamIndex: initialKeyStreamIndex,
        );
      });

      test("first state initialization", () {
        final state = converter.initialState;
        expect(_hexFrom(state),
            equals(_hexFrom(expectedFirstBlockInitialization)));
      });

      test("second state initialization", () {
        final state = new Uint32List.fromList(converter.initialState);
        state[12] += 1;
        expect(_hexFrom(state),
            equals(_hexFrom(expectedSecondBlockInitialization)));
      });

      test("first state after encrypting", () {
        final state = new Uint32List(16);
        converter.encryptState(state);
        expect(_hexFrom(state), equals(_hexFrom(expectedFirstStateEncrypted)));
      });

      test("second state after encrypting", () {
        converter.keyStreamIndex += 64;
        final state = new Uint32List(16);
        converter.encryptState(state);
        expect(_hexFrom(state), equals(_hexFrom(expectedSecondStateEncrypted)));
      });

      test("keystream", () {
        final keyStream = new Uint8List(cleartext.length);
        converter.fillWithKeyStream(keyStream, 0);
        expect(_hexFrom(keyStream), equals(_hexFrom(expectedKeyStream)));
      });

      test("convert", () {
        final encrypted = converter.convert(cleartext);
        expect(_hexFrom(encrypted), equals(_hexFrom(expectedEncrypted)));
      });

      test("convert encrypted to cleartext", () {
        final decrypted = converter.convert(expectedEncrypted);
        expect(_hexFrom(decrypted), equals(_hexFrom(cleartext)));
      });

      test("convert in multiple parts", () {
        // Span 0:0
        var encrypted = converter.convert(cleartext.sublist(0, 0));
        expect(_hexFrom(encrypted),
            equals(_hexFrom(expectedEncrypted.sublist(0, 0))));

        // Span 0:1
        encrypted = converter.convert(cleartext.sublist(0, 1));
        expect(_hexFrom(encrypted),
            equals(_hexFrom(expectedEncrypted.sublist(0, 1))));

        // Span 1:3
        encrypted = converter.convert(cleartext.sublist(1, 3));
        expect(_hexFrom(encrypted),
            equals(_hexFrom(expectedEncrypted.sublist(1, 3))));

        // Span 3:end
        encrypted = converter.convert(cleartext.sublist(3));
        expect(_hexFrom(encrypted),
            equals(_hexFrom(expectedEncrypted.sublist(3))));
      });

      test("fillWithConverted throws ArgumentError if length is negative", () {
        final buffer = new Uint8List(65);
        expect(
          () =>
              converter.fillWithConverted(buffer, 0, cleartext, 0, length: -1),
          throwsA(const TypeMatcher<ArgumentError>()),
        );
      });

      test("fillWithConverted throws ArgumentError if length is too large", () {
        final buffer = new Uint8List(1);
        expect(
          () => converter.fillWithConverted(buffer, 0, cleartext, 0, length: 2),
          throwsA(const TypeMatcher<ArgumentError>()),
        );
      });

      test("fillWithConverted in multiple parts", () {
        final buffer = new Uint8List(65);
        var filledLength = 0;

        // Fill buffer with encrypted 1..3
        converter.keyStreamIndex = initialKeyStreamIndex;

        for (var n in [2, 0, 3]) {
          // Fill
          converter.fillWithConverted(
            buffer,
            filledLength,
            cleartext,
            filledLength,
            length: n,
          );
          filledLength += n;

          // Check
          expect(
            _hexFrom(buffer.sublist(0, filledLength)),
            equals(_hexFrom(expectedEncrypted.sublist(0, filledLength))),
          );

          // Zeroes after it
          expect(
            buffer.sublist(filledLength).every((item) => item == 0),
            isTrue,
          );
        }
      });

      test("fillWithConverted when buffer is smaller than cleartext", () {
        final buffer = new Uint8List(5);
        final bufferStart = 4;
        const clearTextStart = 2;

        // Fill buffer with encrypted 1..3
        converter.keyStreamIndex = initialKeyStreamIndex + clearTextStart;
        converter.fillWithConverted(
            buffer, bufferStart, cleartext, clearTextStart);

        // 0:4 is zeroes
        expect(
            buffer.sublist(0, bufferStart).every((item) => item == 0), isTrue);

        // 4:N is converted
        expect(
          _hexFrom(buffer.sublist(bufferStart)),
          equals(_hexFrom(expectedEncrypted.sublist(
              clearTextStart, clearTextStart + buffer.length - bufferStart))),
        );
      });

      test("fillWithConverted when buffer is larger than cleartext", () {
        final buffer = new Uint8List(512);
        final bufferStart = 4;
        const clearTextStart = 2;

        converter.keyStreamIndex = initialKeyStreamIndex + clearTextStart;
        converter.fillWithConverted(
            buffer, bufferStart, cleartext, clearTextStart);

        // 0:4 is zeroes
        expect(
            buffer.sublist(0, bufferStart).every((item) => item == 0), isTrue);

        // 4:N is converted
        final bufferFilledEnd =
            bufferStart + expectedEncrypted.length - clearTextStart;
        expect(
          _hexFrom(buffer.sublist(bufferStart, bufferFilledEnd)),
          equals(_hexFrom(expectedEncrypted.sublist(clearTextStart))),
        );

        // N: is zeroes
        expect(
            buffer.sublist(bufferFilledEnd).every((item) => item == 0), isTrue);
      });
    });
  });
}

/// Example: [0x0A, 0xABCD] --> "A ABCD"
String _hexFrom(Iterable<int> list) {
  return list.map((v) => v.toRadixString(16)).join(" ");
}

/// Example: ["0a 0b:0c"] --> [10,11,12]
Uint8List _bytesFromHex(String s) {
  return new Uint8List.fromList(s
      .replaceAll("\n", " ")
      .replaceAll(":", " ")
      .split(" ")
      .where((s) => s != "")
      .map((s) => int.parse(s, radix: 16))
      .toList());
}

/// Example: "ABCDEF01 ABCDEF01:ABCDEF01" --> [0xABCDEF01, 0xABCDEF01, 0xABCDEF01]
Uint32List _uint32ListFromHex(String s) {
  return new Uint32List.fromList(s
      .replaceAll("\n", " ")
      .replaceAll(":", " ")
      .split(" ")
      .where((s) => s != "")
      .map((s) => int.parse(s, radix: 16))
      .toList());
}
